/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      ToolBar
Version:   1.0
ID:        ID_TOOLBAR
Status:    Beta
Category:  GUI
Date:      November 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2003.  All rights reserved.
Keywords:  tool, bar, toolbar, toolbars, set, group
Short:     The ToolBar class is used to create and manage groups of toolbars.

###DESCRIPTION###
<p>The ToolBar class creates and manages the use of toolbars in application
interfaces.  A toolbar is a horizontal or vertical strip of items, each of
which performs a specific function.  Each item is presented as a button that
contain an icon image, text string or combination of both.</p>

<p>The definitions for the toolbar are loaded by default from the
environment file "templates:toolbardef.xml".  You can change the template
file prior to initialisation by setting the Template field.  Note that any
values set in the template will override your original field settings for
the toolbar.</p>

<p>Items can be added to the toolbar using XML (recommended for scripting)
or the <a href="#AddToolItem()">AddToolItem</a> method.  Once you have built the toolbar structure, it will
become functional as soon as you display it.  The following example
illustrates how you might create a toolbar using DML code:</p>

<pre>
  &lt;toolbar x="[owner.leftmargin]" y="[owner.topmargin]" xoffset="[owner.rightmargin]"
    height="24"&gt;
    &lt;xml-translate&gt;
      &lt;item text="Cut" tip="Cut" icon="tools/cut"&gt;
        &lt;action object="[fileview]" method="cutfiles"/&gt;
      &lt;/item&gt;

      &lt;item text="Copy" tip="Copy" icon="tools/copy"&gt;
        &lt;action object="[fileview]" method="copyfiles"/&gt;
      &lt;/item&gt;

      &lt;item text="Paste" tip="Paste" icon="tools/paste"&gt;
        &lt;action object="[fileview]" method="pastefiles"/&gt;
      &lt;/item&gt;
    &lt;/xml-translate&gt;
  &lt;/toolbar&gt;
</pre>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/all.h>
#include <pandora/misc/all.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/render.h>
#include <pandora/input/joyport.h>
#include <pandora/input/keyboard.h>
#include <pandora/data/xml.h>
#include <pandora/functions/set.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_TOOLBAR  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_TOOLBAR, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003. All rights reserved.",
   "November 2003",
   "ToolBar"
};

static struct KernelBase *KernelBase;
static struct StringsBase *StringsBase;
static struct RenderBase *RenderBase;
static struct FileSystemBase *FileSystemBase;
static OBJECTPTR ToolBarClass = NULL;
static OBJECTPTR StringsMod = NULL;
static OBJECTPTR FileSystemMod = NULL;
static OBJECTPTR RenderMod = NULL;

#define HGAP 4

enum {
   STATE_ENTERED=1,
   STATE_EXITED,
   STATE_INSIDE
};

/*****************************************************************************
** Object structure.
*/

struct ToolItem {
   UBYTE Text[30];
   struct Picture *Picture;
   STRING Tip;
   STRING Script;
   WORD XCoord;
   WORD YCoord;
   WORD Width;
   WORD Height;
   BYTE Flags;
};

#define TIF_DISABLED  0x01 /* Item is disabled */
#define TIF_STATEMENT 0x02 /* The script reference is a statement string */

struct ToolBar {
   OBJECT_HEADER              /* Class header */
   OBJECTID RegionID;         /* Drawable region created by the toolbar object */
   OBJECTID DrawableID;       /* The drawable target for the toolbar graphic */
   OBJECTID InputID;          /* Keyboard object to use for receiving input */
   LONG     Flags;            /* Special options */
   struct   Font *Font;       /* Font control object */
   LONG     Thickness;        /* Border thickness */
   LONG     Total;            /* Total number of items */
   LONG     Index;            /* Index of the toolbar item that is currently selected */
   FLOAT    Delay;            /* Delay for tool tips */

   PRIVATE_FIELDS
   struct   RGB HighlightRGB; /* Colour to use for border highlighting */
   struct   RGB ShadowRGB;    /* Colour to use for border shadowing */
   OBJECTPTR TemplateXML;
   OBJECTPTR XML;
   OBJECTID FontID;
   BYTE     KeyboardChannel;
   BYTE     State;
   BYTE     Clicked;          /* Set to TRUE if the toolbar has been clicked.  Reverts back to FALSE when the user releases the toolbar */
   BYTE     Active;
   WORD     ClickX;
   WORD     ClickY;
   LONG     DrawableWidth;
   LONG     DrawableHeight;
   LONG     MotionTimer;
   OBJECTID ToolTipID;
   struct ToolItem Items[40];
};

#define TBF_HIDE       0x00000001 /* Hide on creation, or toolbar is currently hidden*/
#define TBF_DISABLED   0x00000002 /* Disable on creation */
#define TBF_NOBKGD     0x00000008 /* Do not draw background graphics (automatically set if Template graphics are used) */
#define TBF_LOCAL      0x00000010 /* On activation, execute objects locally rather than messaging them */
#define TBF_SUNKEN     0x00000040 /* Reverses the highlight and shadow definitions to create a sunken effect */
#define TBF_EMBOSSED   0x00000080 /* The border effect is embossed */

/*****************************************************************************
** Class definition.
*/

static struct FieldDef ToolBarFlags[] = {
   { "HIDE",       TBF_HIDE },
   { "DISABLED",   TBF_DISABLED },
   { "NOBKGD",     TBF_NOBKGD },
   { "LOCAL",      TBF_LOCAL },
   { "SUNKEN",     TBF_SUNKEN },
   { "EMBOSSED",   TBF_EMBOSSED },
   { NULL, NULL }
};

static ERROR GET_Bottom(struct ToolBar *, LONG *);
static ERROR GET_Height(struct ToolBar *, struct Variable *);
static ERROR GET_HighlightRGB(struct ToolBar *, struct RGB *);
static ERROR GET_Right(struct ToolBar *, LONG *);
static ERROR GET_ShadowRGB(struct ToolBar *, struct RGB *);
static ERROR GET_Width(struct ToolBar *, struct Variable *);
static ERROR GET_X(struct ToolBar *, struct Variable *);
static ERROR GET_XOffset(struct ToolBar *, struct Variable *);
static ERROR GET_Y(struct ToolBar *, struct Variable *);
static ERROR GET_YOffset(struct ToolBar *, struct Variable *);

static ERROR SET_Height(struct ToolBar *, struct Variable *);
static ERROR SET_Highlight(struct ToolBar *, STRING);
static ERROR SET_Shadow(struct ToolBar *, STRING);
static ERROR SET_TabFocus(struct ToolBar *, OBJECTID *);
static ERROR SET_Template(struct ToolBar *, STRING);
static ERROR SET_Width(struct ToolBar *, struct Variable *);
static ERROR SET_X(struct ToolBar *, struct Variable *);
static ERROR SET_XOffset(struct ToolBar *, struct Variable *);
static ERROR SET_Y(struct ToolBar *, struct Variable *);
static ERROR SET_YOffset(struct ToolBar *, struct Variable *);

static struct FieldArray ToolBarFields[] = {
   { "Region",         0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Drawable",       0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Input",          0, FDF_OBJECTID|FDF_RI,  ID_KEYBOARD, NULL ,NULL },
   { "Flags",          0, FDF_LONGFLAGS|FDF_RW, (LONG)&ToolBarFlags, NULL, NULL },
   { "Font",           0, FDF_CHILD|FDF_R,      0, NULL, NULL },
   { "Thickness",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Total",          0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Index",          0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Delay",          0, FDF_FLOAT|FDF_RW,     0, NULL, NULL },
   /** Virtual fields ***/
   { "Bottom",         0, FDF_LONG|FDF_R,      0, GET_Bottom,       NULL },
   { "Highlight",      0, FDF_STRING|FDF_W,    0, NULL,             SET_Highlight },
   { "HighlightRGB",   0, FDF_POINTER|FDF_R,   0, GET_HighlightRGB, NULL },
   { "Right",          0, FDF_LONG|FDF_R,      0, GET_Right,        NULL },
   { "Shadow",         0, FDF_STRING|FDF_W,    0, NULL,             SET_Shadow },
   { "ShadowRGB",      0, FDF_POINTER|FDF_R,   0, GET_ShadowRGB,    NULL },
   { "TabFocus",       0, FDF_OBJECTID|FDF_W,  ID_TABFOCUS, NULL,   SET_TabFocus },
   { "Template",       0, FDF_STRING|FDF_I,    0, NULL,             SET_Template },
   /*** Variable Fields ***/
   { "Height",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",          0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR TOOLBAR_AccessObject(struct ToolBar *, APTR);
static ERROR TOOLBAR_ActionNotify(struct ToolBar *, struct acActionNotify *);
static ERROR TOOLBAR_Activate(struct ToolBar *, APTR);
static ERROR TOOLBAR_ClosingTag(struct ToolBar *, APTR);
static ERROR TOOLBAR_DataChannel(struct ToolBar *, struct acDataChannel *);
static ERROR TOOLBAR_Disable(struct ToolBar *, APTR);
static ERROR TOOLBAR_Enable(struct ToolBar *, APTR);
static ERROR TOOLBAR_Focus(struct ToolBar *, APTR);
static ERROR TOOLBAR_Free(struct ToolBar *, APTR);
static ERROR TOOLBAR_Hide(struct ToolBar *, APTR);
static ERROR TOOLBAR_LostFocus(struct ToolBar *, APTR);
static ERROR TOOLBAR_Init(struct ToolBar *, APTR);
static ERROR TOOLBAR_MoveToBack(struct ToolBar *, APTR);
static ERROR TOOLBAR_MoveToFront(struct ToolBar *, APTR);
static ERROR TOOLBAR_NewObject(struct ToolBar *, APTR);
static ERROR TOOLBAR_Redimension(struct ToolBar *, struct acRedimension *);
static ERROR TOOLBAR_ReleaseObject(struct ToolBar *, APTR);
static ERROR TOOLBAR_Resize(struct ToolBar *, struct acResize *);
static ERROR TOOLBAR_Show(struct ToolBar *, APTR);
static ERROR TOOLBAR_Timer(struct ToolBar *, struct acTimer *);
static ERROR TOOLBAR_UserClick(struct ToolBar *, struct acUserClick *);
static ERROR TOOLBAR_UserClickRelease(struct ToolBar *, struct acUserClickRelease *);
static ERROR TOOLBAR_UserMovement(struct ToolBar *, struct acUserMovement *);

static struct ActionArray ToolBarActions[] = {
   { AC_AccessObject,     TOOLBAR_AccessObject },
   { AC_ActionNotify,     TOOLBAR_ActionNotify },
   { AC_Activate,         TOOLBAR_Activate },
   { AC_ClosingTag,       TOOLBAR_ClosingTag },
   { AC_DataChannel,      TOOLBAR_DataChannel },
   { AC_Disable,          TOOLBAR_Disable },
   { AC_Enable,           TOOLBAR_Enable },
   { AC_Focus,            TOOLBAR_Focus },
   { AC_Free,             TOOLBAR_Free },
   { AC_Hide,             TOOLBAR_Hide },
   { AC_Init,             TOOLBAR_Init },
   { AC_LostFocus,        TOOLBAR_LostFocus },
   { AC_MoveToBack,       TOOLBAR_MoveToBack },
   { AC_MoveToFront,      TOOLBAR_MoveToFront },
   { AC_NewObject,        TOOLBAR_NewObject },
   { AC_Redimension,      TOOLBAR_Redimension },
   { AC_ReleaseObject,    TOOLBAR_ReleaseObject },
   { AC_Resize,           TOOLBAR_Resize },
   { AC_Show,             TOOLBAR_Show },
   { AC_Timer,            TOOLBAR_Timer },
   { AC_UserClick,        TOOLBAR_UserClick },
   { AC_UserClickRelease, TOOLBAR_UserClickRelease },
   { AC_UserMovement,     TOOLBAR_UserMovement },
   { NULL, NULL }
};

/*****************************************************************************
** Class methods.
*/

static ERROR TOOLBAR_AddItem(struct ToolBar *, struct mtAddToolItem *);
static ERROR TOOLBAR_DisableItem(struct ToolBar *, struct mtDisableToolItem *);
static ERROR TOOLBAR_EnableItem(struct ToolBar *, struct mtEnableToolItem *);
static ERROR TOOLBAR_RemoveItem(struct ToolBar *, struct mtRemoveToolItem *);

static struct FunctionField argsAddItem[]     = { { "Name", ARG_STR }, { "Tip", ARG_STR }, { "Icon", ARG_STR }, { "Script", ARG_STR }, { NULL, NULL } };
static struct FunctionField argsDisableItem[] = { { "Name", ARG_STR }, { NULL, NULL } };
static struct FunctionField argsEnableItem[]  = { { "Name", ARG_STR }, { NULL, NULL } };
static struct FunctionField argsRemoveItem[]  = { { "Name", ARG_STR }, { NULL, NULL } };

static struct MethodArray ToolBarMethods[] = {
   { MT_AddToolItem,     TOOLBAR_AddItem,     "AddItem",     argsAddItem,     sizeof(struct mtAddToolItem) },
   { MT_RemoveToolItem,  TOOLBAR_RemoveItem,  "RemoveItem",  argsRemoveItem,  sizeof(struct mtRemoveToolItem) },
   { MT_DisableToolItem, TOOLBAR_DisableItem, "DisableItem", argsDisableItem, sizeof(struct mtDisableToolItem) },
   { MT_EnableToolItem,  TOOLBAR_EnableItem,  "EnableItem",  argsEnableItem,  sizeof(struct mtEnableToolItem) },
   { NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Prototypes
*/

static void ArrangeItems(struct ToolBar *);
static void DrawToolBar(struct ToolBar *, struct Render *, struct Bitmap *);
static void DrawItem(struct ToolBar *, WORD);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL, FID_Name|TSTRING, "strings", TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &FileSystemMod, NULL, FID_Name|TSTRING, "FileSystem", TAGEND) IS ERR_Okay) {
      if (GetField(FileSystemMod, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &RenderMod, NULL, FID_Name|TSTRING, "Render", TAGEND) IS ERR_Okay) {
      if (GetField(RenderMod, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ToolBarClass, NULL,
      FID_BaseClassID|TLONG, ID_TOOLBAR,
      FID_Version|TFLOAT,    VER_TOOLBAR,
      FID_Name|TSTRING,      "ToolBar",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN|CLF_PRIVATEONLY,
      FID_Actions|TPTR,      ToolBarActions,
      FID_Methods|TPTR,      ToolBarMethods,
      FID_Fields|TPTR,       ToolBarFields,
      FID_Size|TLONG,        sizeof(struct ToolBar),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (ToolBarClass)  { Action(AC_Free, ToolBarClass, NULL);  ToolBarClass  = NULL; }
   if (StringsMod)    { Action(AC_Free, StringsMod, NULL);    StringsMod    = NULL; }
   if (FileSystemMod) { Action(AC_Free, FileSystemMod, NULL); FileSystemMod = NULL; }
   if (RenderMod)     { Action(AC_Free, RenderMod, NULL);     RenderMod     = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** ToolBar: AccessObject
*/

static ERROR TOOLBAR_AccessObject(struct ToolBar *Self, APTR Void)
{
   if ((Self->FontID) AND (!Self->Font)) {
      if (AccessObject(Self->FontID, 0, (OBJECTPTR *)&Self->Font) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************
** ToolBar: ActionNotify
*/

static ERROR TOOLBAR_ActionNotify(struct ToolBar *Self, struct acActionNotify *Args)
{
   if (Args->ActionID IS AC_Focus) {
      OBJECTPTR input;

      DPrintF("~Focus()","[ToolBar:%d]", Self->Head.UniqueID);

      if (Self->KeyboardChannel IS FALSE) {
         if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
            SubscribeChannel(input, Self);
            Self->KeyboardChannel = TRUE;
            ReleaseObject(input);
         }
         else ObjectError(Self, ERH_Focus, ERR_AccessObject);
      }

      DelayMsg(AC_Draw, Self->RegionID, NULL);

      StepBack();
   }
   else if (Args->ActionID IS AC_Disable) {
      Self->Flags |= TBF_DISABLED;

      if (Self->ToolTipID) {
         ActionMsg(AC_Free, Self->ToolTipID, NULL);
         Self->ToolTipID = NULL;
      }

      UnsubscribeTimer(Self->Head.UniqueID);

      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }
   else if (Args->ActionID IS AC_Enable) {
      Self->Flags &= ~TBF_DISABLED;
      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }
   else if (Args->ActionID IS AC_LostFocus) {
      Action(AC_LostFocus, Self, NULL);
   }
   else if (Args->ActionID IS AC_Redimension) {
      ArrangeItems(Self);
   }
   else if (Args->ActionID IS AC_UserClick) {
      Action(AC_UserClick, Self, Args->Args);
   }
   else if (Args->ActionID IS AC_UserClickRelease) {
      Action(AC_UserClickRelease, Self, Args->Args);
   }
   else if (Args->ActionID IS AC_UserMovement) {
      Action(AC_UserMovement, Self, Args->Args);
   }
   else return(ERR_NoSupport);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Activates the currently selected item in the toolbar.
###END###

*****************************************************************************/

static ERROR TOOLBAR_Activate(struct ToolBar *Self, APTR Void)
{
   OBJECTPTR script;

   DPrintF("~Activate()","[ToolBar:%d] Item: %d", Self->Head.UniqueID, Self->Index);

   if (Self->Active IS TRUE) {
      DPrintF("@Activate:","[ToolBar] Warning - recursion detected");
      StepBack();
      return(ERR_Failed);
   }

   Self->Active = TRUE;

   if ((Self->Index >= 0) AND (Self->Index < Self->Total)) {
      if (Self->Items[Self->Index].Script) {
         if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
            FID_Statement|TSTRING, Self->Items[Self->Index].Script,
            TAGEND) IS ERR_Okay) {
            Action(AC_Activate, script, NULL);
            Action(AC_Free, script, NULL);
         }
      }
   }

   Self->Active = FALSE;

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    AddToolItem
Synonym: AddItem
Short:   Adds a new item to the toolbar.

###DESCRIPTION###
<p>New items can be added to the toolbar by calling this method.  At a minimum you are
required to specify a Name for the new item.  The Name will
double-up as the item text that is displayed in the toolbar.  Set the Tip string
if you would like a tool-tip to be displayed when the mouse pointer is positioned
over the item.</p>

<p>The Icon parameter allows icons to be displayed instead of text inside
the toolbar.  The icon string should reference an image from the icon
database, using the format "category/icon".  You may specify a custom icon
image if you wish - the routine will detect this if it determines that you
have specified a complete file path in the Icon string.</p>

<p>To execute a script <i>statement</i> when the user clicks on the item, set the
Script field.  The string that you provide must be in a recognisable script format, such as
DML code.  If you would like to execute a script <i>file</i> when the item is
activated, use the string format 'script src="path:file"'.  Note: Your script must not
create static objects that could linger after the script has been executed.</p>

###INPUT###
STRING/Name: The name or text to display for the item.
STRING/Tip: A tool-tip to display for the item when the user's pointer hovers over it.
STRING/Icon: The icon to use for the item.
STRING/Script: A script statement to be executed when the item is clicked by the user.

###RESULT###
ERR_Okay: The new item was added successfully.
ERR_Args: Invalid arguments (the Name must be specified at a minimum).
ERR_ArrayFull: No more items can be added because the toolbar is at maximum capacity.

###END###

*****************************************************************************/

static ERROR TOOLBAR_AddItem(struct ToolBar *Self, struct mtAddToolItem *Args)
{
   struct Picture *picture;
   UBYTE buffer[300];
   WORD i, index;

   if ((!Args) OR (!Args->Name)) return(ObjectError(Self, MT_AddToolItem, ERR_Args));

   DPrintF("~AddItem()","[ToolBar] Name: %s, Icon: %s %s", Args->Name, Args->Icon, Args->Script);

   if (Self->Total >= ARRAYSIZE(Self->Items)) return(StepObjectError(Self, MT_AddToolItem, ERR_ArrayFull));

   index = Self->Total;
   ClearMemory(Self->Items + index, sizeof(Self->Items[0]));

   StrCopy(Args->Name, Self->Items[index].Text, sizeof(Self->Items[index].Text));

   if (Args->Tip) {
      if (AllocMemory(StrLength(Args->Tip) + 1, MEM_STRING, &Self->Items[index].Tip, NULL) IS ERR_Okay) {
         StrCopy(Args->Tip, Self->Items[index].Tip, COPY_ALL);
      }
   }

   if (Args->Icon) {
      /*** Load the icon file as a picture ***/

      if (StrCompare("icons:", Args->Icon, 6, NULL) IS ERR_Okay) {
         StrFormat(buffer, sizeof(buffer), "%s(16)", Args->Icon);
      }
      else {
         for (i=0; (Args->Icon[i]) AND (Args->Icon[i] != ':'); i++);
         if (Args->Icon[i] IS ':') StrCopy(Args->Icon, buffer, sizeof(buffer));
         else StrFormat(buffer, sizeof(buffer), "icons:%s(16)", Args->Icon);
      }

      if (NewObject(ID_PICTURE, NF_CHILD, &picture, NULL) IS ERR_Okay) {
         SetFields(picture, FID_Location|TSTRING, buffer, TAGEND);
         picture->Flags |= PCF_32BITALPHA|PCF_CACHE;

         if (Action(AC_Init, picture, NULL) IS ERR_Okay) {
            Self->Items[index].Picture = picture;
         }
         else {
            Action(AC_Free, picture, NULL);
            picture = NULL;
            DPrintF("@AddItem:","[ToolBar] Failed to load toolbar icon.");
         }
      }
      else ObjectError(Self, MT_AddToolItem, ERR_NewObject);
   }

   if (Args->Script) {
      if (AllocMemory(StrLength(Args->Script) + 1, MEM_STRING, &Self->Items[index].Script, NULL) IS ERR_Okay) {
         StrCopy(Args->Script, Self->Items[index].Script, COPY_ALL);
      }
   }

   Self->Total++;

   ArrangeItems(Self);

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** ToolBar: ClosingTag
*/

static ERROR TOOLBAR_ClosingTag(struct ToolBar *Self, APTR Void)
{
   if (!(Self->Flags & TBF_HIDE)) Action(AC_Show, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** ToolBar: DataChannel
*/

static ERROR TOOLBAR_DataChannel(struct ToolBar *Self, struct acDataChannel *Args)
{
   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      struct dcKeyEntry *key;
      WORD i, oldindex;

      if ((key = (struct dcKeyEntry *)Args->Buffer)) {
         for (i=0; i < Args->TotalEntries; i++) {
            if (key[i].Flags & KQ_PRESSED) {
               if ((key[i].Value IS K_ENTER) OR (key[i].Value IS K_NPENTER) OR (key[i].Value IS K_SPACE)) {
                  Action(AC_Activate, Self, NULL);
               }
               else if ((key[i].Value IS K_UP) OR (key[i].Value IS K_LEFT)) {
                  if (Self->Index != 0) {
                     oldindex = Self->Index;
                     Self->Index--;
                     while ((Self->Index > 0) AND (StrCompare("BREAK", Self->Items[Self->Index].Text, 0, STR_MATCHLENGTH) IS ERR_Okay)) Self->Index--;
                     if (Self->Index < 0) Self->Index = 0;
                     if (StrCompare("BREAK", Self->Items[i].Text, 0, STR_MATCHLENGTH) IS ERR_Okay) Self->Index = -1;

                     DrawItem(Self, oldindex);
                     DrawItem(Self, Self->Index);
                  }
               }
               else if ((key[i].Value IS K_DOWN) OR (key[i].Value IS K_RIGHT)) {
                  oldindex = Self->Index;
                  Self->Index++;
                  while ((Self->Index < Self->Total) AND (StrCompare("BREAK", Self->Items[Self->Index].Text, 0, STR_MATCHLENGTH) IS ERR_Okay)) Self->Index++;
                  if (Self->Index >= Self->Total) Self->Index = Self->Total-1;
                  DrawItem(Self, oldindex);
                  DrawItem(Self, Self->Index);
               }
            }
         }
      }
   }
   else if ((Args->DataType IS DATA_XML) AND (Args->Version IS 1)) {
      struct XMLTag **xmllist, *tag;
      struct mtAddToolItem add;
      struct mtGetXMLString getxml;
      LONG n, itemcount;

      if (!Self->XML) {
         if (CreateObject(ID_XML, NF_CHILD, &Self->XML, NULL,
            FID_Statement|TSTRING, Args->Buffer,
            NULL) != ERR_Okay) {
            return(ObjectError(Self, ERH_DataChannel, ERR_CreateObject));
         }
      }
      else if (SetField(Self->XML, FID_Statement, FT_POINTER, Args->Buffer) != ERR_Okay) {
         return(ERR_SetField);
      }

      if (GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) {

         /*** Search for <item> tags and use them ***/

         itemcount = 0;
         for (tag=xmllist[0]; tag; tag=tag->Next) {
            if (StrCompare("item", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {

               ClearMemory(&add, sizeof(add));

               for (n=0; n < tag->TotalAttributes; n++) {
                  if (StrCompare("text", tag->Attributes[n].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     add.Name = tag->Attributes[n].Value;
                  }
                  else if (StrCompare("tip", tag->Attributes[n].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     add.Tip = tag->Attributes[n].Value;
                  }
                  else if (StrCompare("icon", tag->Attributes[n].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     add.Icon = tag->Attributes[n].Value;
                  }
               }

               if (tag->Child) {
                  getxml.Index = tag->Child->Index;
                  getxml.Flags = NULL;
                  Action(MT_GetXMLString, Self->XML, &getxml);

                  add.Script = getxml.Result;
               }

               Action(MT_AddToolItem, Self, &add);

               if (add.Script) FreeMemory(add.Script);
            }
            else if (StrCompare("break", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               add.Name   = "BREAK";
               add.Tip    = NULL;
               add.Icon   = NULL;
               add.Script = NULL;
               Action(MT_AddToolItem, Self, &add);
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Disable
Short: Turns the entire toolbar off.
###END###

*****************************************************************************/

static ERROR TOOLBAR_Disable(struct ToolBar *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is disabled.
   */

   DPrintF("Disable()","[ToolBar:%d]", Self->Head.UniqueID);

   ActionMsg(AC_Disable, Self->RegionID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  DisableItem
Short: Disables a toolbar item.

###DESCRIPTION###
<p>Individual toolbar items can be disabled by calling this method.  You are
required to provide the name of the item that you want to disable.  If
multiple items share the same name, they will all be disabled.  The toolbar
will be automatically redrawn as a result of calling this method.</p>

<p>A disabled item can be re-enabled by calling the <a
href="#EnableItem()">EnableItem()</a> method.</p>

###INPUT###
STRING/Name: The name of the item to be disabled.

###RESULT###
ERR_Okay: The item was either disabled or does not exist.
ERR_Args: A valid Name was not specified.

###END###

*****************************************************************************/

static ERROR TOOLBAR_DisableItem(struct ToolBar *Self, struct mtDisableToolItem *Args)
{
   WORD i;

   if ((!Args) OR (!Args->Name) OR (!Args->Name[0])) return(ERR_Args);

   for (i=0; i < Self->Total; i++) {
      if (StrCompare(Args->Name, Self->Items[i].Text, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if (!(Self->Items[i].Flags & TIF_DISABLED)) {
            Self->Items[i].Flags |= TIF_DISABLED;
            DrawItem(Self, i);
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  EnableItem
Short: Enables a toolbar item that has been earlier disabled.

###DESCRIPTION###
<p>Items that have been disabled can be re-enabled by calling this method.
You are required to provide the name of the item that you want to enable.
If multiple items share the same name, they will all be enabled.</p>

###INPUT###
STRING/Name: The name of the item to be enabled.

###RESULT###
ERR_Okay: The item was either enabled or does not exist.
ERR_Args: A valid Name was not specified.

###END###

*****************************************************************************/

static ERROR TOOLBAR_EnableItem(struct ToolBar *Self, struct mtEnableToolItem *Args)
{
   WORD i;

   if ((!Args) OR (!Args->Name) OR (!Args->Name[0])) return(ERR_Args);

   for (i=0; i < Self->Total; i++) {
      if (StrCompare(Args->Name, Self->Items[i].Text, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if (Self->Items[i].Flags & TIF_DISABLED) {
            Self->Items[i].Flags &= ~TIF_DISABLED;
            DrawItem(Self, i);
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Enable
Short: Turns the toolbar on if it has been disabled.
###END###

*****************************************************************************/

static ERROR TOOLBAR_Enable(struct ToolBar *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is enabled.
   */
   DPrintF("Enable()","[ToolBar:%d]", Self->Head.UniqueID);
   ActionMsg(AC_Enable, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Sets the focus on the toolbar and activates keyboard monitoring.
###END###

*****************************************************************************/

static ERROR TOOLBAR_Focus(struct ToolBar *Self, APTR Void)
{
   return(ActionMsg(AC_Focus, Self->RegionID, NULL));
}

/*****************************************************************************
** ToolBar: Free
*/

static ERROR TOOLBAR_Free(struct ToolBar *Self, APTR Void)
{
   WORD i;

   UnsubscribeTimer(Self->Head.UniqueID);

   for (i=0; i < Self->Total; i++) {
      if (Self->Items[i].Script) {
         FreeMemory(Self->Items[i].Script);
         Self->Items[i].Script = NULL;
      }

      if (Self->Items[i].Tip) {
         FreeMemory(Self->Items[i].Tip);
         Self->Items[i].Tip = NULL;
      }

      if (Self->Items[i].Picture) {
         Action(AC_Free, Self->Items[i].Picture, NULL);
         Self->Items[i].Picture = NULL;
      }
   }

   if (Self->XML) {
      Action(AC_Free, Self->XML, NULL);
      Self->XML = NULL;
   }

   if (Self->FontID) {
      if (Self->Font) {
         Action(AC_Free, Self->Font, NULL);
         ReleaseObject(Self->Font);
         Self->Font = NULL;
      }
      else ActionMsg(AC_Free, Self->FontID, NULL);
      Self->FontID = NULL;
   }

   if (Self->RegionID) {
      ActionMsg(AC_Free, Self->RegionID, NULL);
      Self->RegionID = NULL;
   }

   if (Self->TemplateXML) {
      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Removes the toolbar from the display.
###END###

*****************************************************************************/

static ERROR TOOLBAR_Hide(struct ToolBar *Self, APTR Void)
{
   ActionMsg(AC_Hide, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** ToolBar: Init()
*/

static ERROR TOOLBAR_Init(struct ToolBar *Self, APTR Void)
{
   struct Render *region;
   OBJECTID owner_id;

   /*** Find our parent drawable ***/

   if (!Self->DrawableID) {
      owner_id = GetOwner(Self);
      while ((owner_id) AND (GetClassID(owner_id) != ID_RENDER)) {
         owner_id = GetOwnerID(owner_id);
      }
      if (owner_id) Self->DrawableID = owner_id;
      else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
   }

   /*** Setup keyboard ***/

   if (!Self->InputID) {
      if (FastFindObject("SystemKeyboard", NULL, &Self->InputID, 1, NULL) != ERR_Okay) {
         DPrintF("@Init:","[ToolBar] Warning: Could not find any SystemKeyboard objects.");
      }
   }

   /*** Initialise the Font ***/

   if (Action(AC_Init, Self->Font, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   /*** Initialise the toolbar region ***/

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      SetFields(region, FID_Parent|TLONG, Self->DrawableID,
                        FID_Region|TLONG, TRUE,
                        TAGEND);

      if (!(region->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if ((!(region->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y))) OR (!(region->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)))) {
            SetFields(region, FID_Height|TLONG, 24, TAGEND);
         }
      }

      if (Action(AC_Init, region, NULL) IS ERR_Okay) {
         SubscribeActionTags(region, Self,
            AC_Disable,
            AC_Enable,
            AC_Focus,
            AC_LostFocus,
            AC_Redimension,
            AC_UserClick,
            AC_UserClickRelease,
            AC_UserMovement,
            TAGEND);

         Self->DrawableWidth = region->Width;
         Self->DrawableHeight = region->Height;
      }

      ReleaseObject(region);
   }
   else return(ERR_AccessObject);

   /*** Use the base template to create the toolbar graphics ***/

   if (Self->TemplateXML) {
      OBJECTPTR script;
      struct mtGetXMLString getxml;
      struct XMLTag **xmllist, *tag;

      if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
         for (tag=xmllist[0]; tag; tag=tag->Next) {
            if (tag->Child) {
               if (StrCompare("graphics", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  getxml.Index = tag->Child->Index;
                  getxml.Flags = XMF_INCLUDESIBLINGS;
                  Action(MT_GetXMLString, Self->TemplateXML, &getxml);

                  if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
                     FID_String|TSTRING, getxml.Result,
                     FID_Target|TLONG,   Self->RegionID,
                     TAGEND) IS ERR_Okay) {

                     if (Action(AC_Activate, script, NULL) IS ERR_Okay) {
                        Self->Flags |= TBF_NOBKGD;
                     }
                  }

                  FreeMemory(getxml.Result);
                  break;
               }
            }
         }
      }

      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, region, Self, &DrawToolBar);
      ReleaseObject(region);
   }
   else return(ERR_AccessObject);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  LostFocus
Short: Removes the focus from the toolbar and disables key monitoring.
###END###

*****************************************************************************/

static ERROR TOOLBAR_LostFocus(struct ToolBar *Self, APTR Void)
{
   OBJECTPTR input;

   if (Self->KeyboardChannel IS TRUE) {
      if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
         UnsubscribeChannel(input, Self->Head.UniqueID);
         Self->KeyboardChannel = FALSE;
         ReleaseObject(input);
      }
      else ObjectError(Self, ERH_LostFocus, ERR_AccessObject);
   }

   if (Self->Index != -1) {
      WORD oldindex;
      oldindex = Self->Index;
      Self->Index = -1;
      DrawItem(Self, oldindex);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToBack
Short: Moves the toolbar to the back of the display area.
###END###

*****************************************************************************/

static ERROR TOOLBAR_MoveToBack(struct ToolBar *Self, APTR Void)
{
   ActionMsg(AC_MoveToBack, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToFront
Short: Moves the toolbar to the front of the display area.
###END###

*****************************************************************************/

static ERROR TOOLBAR_MoveToFront(struct ToolBar *Self, APTR Void)
{
   ActionMsg(AC_MoveToFront, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** ToolBar: NewObject()
*/

static ERROR TOOLBAR_NewObject(struct ToolBar *Self, APTR Void)
{
   if (NewObject(ID_RENDER, NF_CHILD|Self->Head.ObjectFlags, NULL, &Self->RegionID) IS ERR_Okay) {
      if (NewObject(ID_FONT, NF_CHILD|Self->Head.ObjectFlags, &Self->Font, &Self->FontID) IS ERR_Okay) {
         SetField(Self->Font, FID_Face, FT_STRING, "sans serif");

         /*** Shadow colour ***/
         Self->ShadowRGB.Red   = 100;
         Self->ShadowRGB.Green = 100;
         Self->ShadowRGB.Blue  = 100;

         /*** Highlight colour ***/
         Self->HighlightRGB.Red   = 255;
         Self->HighlightRGB.Green = 255;
         Self->HighlightRGB.Blue  = 255;

         Self->Thickness = 2;
         Self->Flags |= TBF_EMBOSSED|TBF_SUNKEN;
         Self->Index = -1;
         Self->Delay = 0.75;

         SET_Template(Self, "templates:toolbardef.xml");

         return(ERR_Okay);
      }
      else return(ERR_NewObject);
   }
   else return(ERR_NewObject);
}

/*****************************************************************************

###ACTION###
Name:  Redimension
Short: Changes the size and position of the toolbar.
###END###

*****************************************************************************/

static ERROR TOOLBAR_Redimension(struct ToolBar *Self, struct acRedimension *Args)
{
   return(ActionMsg(AC_Redimension, Self->RegionID, Args));
}

/*****************************************************************************
** ToolBar: ReleaseObject
*/

static ERROR TOOLBAR_ReleaseObject(struct ToolBar *Self, APTR Void)
{
   if ((Self->Font) AND (Self->FontID)) {
      ReleaseObject(Self->Font);
      Self->Font = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  RemoveItem
Short: Removes an item from the toolbar.

###DESCRIPTION###
<p>This method will remove items from the toolbar.  You need to provide the
name of the item that you want to remove.  If items matching the name
that you provide will be deleted.</p>

<p>The toolbar will be automatically redrawn as a result of calling this
method.</p>

###INPUT###
STRING/Name: The name of the item that you want to remove.

###RESULT###
ERR_Okay: The item was removed.
ERR_Args: A valid Name parameter was not specified.

###END###

*****************************************************************************/

static ERROR TOOLBAR_RemoveItem(struct ToolBar *Self, struct mtRemoveToolItem *Args)
{
   WORD i;

   if ((!Args) OR (!Args->Name) OR (!Args->Name[0])) return(ERR_Args);

   for (i=0; i < Self->Total; i++) {
      if (StrCompare(Args->Name, Self->Items[i].Text, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         CopyMemory(Self->Items + i + 1, Self->Items + i, sizeof(Self->Items[0]) * (Self->Total - i - 1));
         Self->Total--;
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Resize
Short: Alters the size of the toolbar.
###END###

*****************************************************************************/

static ERROR TOOLBAR_Resize(struct ToolBar *Self, struct acResize *Args)
{
   return(ActionMsg(AC_Resize, Self->RegionID, Args));
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Puts the toolbar on display.
###END###

*****************************************************************************/

static ERROR TOOLBAR_Show(struct ToolBar *Self, APTR Void)
{
   ActionMsg(AC_Show, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** ToolBar: Timer
*/

static ERROR TOOLBAR_Timer(struct ToolBar *Self, struct acTimer *Args)
{
   OBJECTPTR script;

   if ((Self->Index >= 0) AND (Self->Index < Self->Total)) {
      if ((Self->MotionTimer < Self->Delay * 1000) AND (Self->MotionTimer + Args->TimeElapsed >= Self->Delay * 1000)) {

         DebugState(FALSE, -1);

         if (Self->ToolTipID) {
            ActionMsg(AC_Free, Self->ToolTipID, NULL);
            Self->ToolTipID = NULL;
         }

         if ((Self->Items[Self->Index].Tip) AND (Self->Items[Self->Index].Tip[0])) {
            if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
                  FID_Location|TSTRING, "templates:tooltip",
                  FID_Text|TUNLISTED,   Self->Items[Self->Index].Tip,
                  TAGEND) IS ERR_Okay) {

               Action(AC_Activate, script, NULL);

               if ((GetField(script, FID_PrimaryObject, FT_LONG, &Self->ToolTipID) != ERR_Okay) OR (!Self->ToolTipID)) {
                  DPrintF("@ToolBar:","[Timer] Failed to retrieve the object hook for the tool tip.");
               }

               Action(AC_Free, script, NULL);
            }
         }

         DebugState(TRUE, -1);
      }

      Self->MotionTimer += Args->TimeElapsed;
   }
   else {
      UnsubscribeTimer(Self->Head.UniqueID);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserClick
Short: Emulates a user click on the toolbar.
###END###

*****************************************************************************/

static ERROR TOOLBAR_UserClick(struct ToolBar *Self, struct acUserClick *Args)
{
   if (!Args) return(ObjectError(Self, ERH_UserClick, ERR_Args));

   if (Self->Flags & TBF_DISABLED) return(ERR_Okay);

   if ((Self->Index >= 0) AND (Self->Index < Self->Total)) {
      if (Self->Items[Self->Index].Flags & TIF_DISABLED) return(ERR_Okay);

      if (Args->ButtonFlags & JD_LMB) {
         Self->Clicked = TRUE;
         Self->ClickX  = Args->XCoord;
         Self->ClickY  = Args->YCoord;

         DrawItem(Self, Self->Index);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserClickRelease
Short: Emulates a user click-release on the toolbar.
###END###

*****************************************************************************/

static ERROR TOOLBAR_UserClickRelease(struct ToolBar *Self, struct acUserClickRelease *Args)
{
   WORD clickx, clicky;

   if (!Args) return(ObjectError(Self, ERH_UserClickRelease, ERR_Args));

   if ((Args->ButtonFlags & JD_LMB) AND (Self->Clicked IS TRUE)) {
      Self->Clicked = FALSE;
      clickx = Args->XCoord - Self->ClickX;
      clicky = Args->YCoord - Self->ClickY;
      if (clickx < 0) clickx = -clickx;
      if (clicky < 0) clicky = -clicky;

      DrawItem(Self, Self->Index);

      if ((clickx < 4) AND (clicky < 4)) {
         Action(AC_Activate, Self, NULL);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserMovement
Short: Emulates user movement over the toolbar.
###END###

*****************************************************************************/

static ERROR TOOLBAR_UserMovement(struct ToolBar *Self, struct acUserMovement *Args)
{
   WORD index, oldindex, i;

   if (!Args) return(ObjectError(Self, ERH_UserMovement, ERR_Args));

   /*** Kill any existing tool-tip due to pointer movement ***/

   if (Self->ToolTipID) {
      ActionMsg(AC_Free, Self->ToolTipID, NULL);
      Self->ToolTipID = NULL;
   }

   Self->MotionTimer = 0;
   UnsubscribeTimer(Self->Head.UniqueID);

   /*** Do nothing if the user is holding the mouse button ***/

   if (Self->Clicked IS TRUE) return(ERR_Okay);

   /*** Determine what item we are positioned over ***/

   index = -1;
   if (Args->OverObject IS Self->RegionID) {
      for (i=0; i < Self->Total; i++) {
         if ((Args->XCoord >= Self->Items[i].XCoord) AND (Args->XCoord <  Self->Items[i].XCoord + Self->Items[i].Width) AND
             (Args->YCoord >= Self->Items[i].YCoord) AND (Args->YCoord <  Self->Items[i].YCoord + Self->Items[i].Height)) {
            if (StrCompare("BREAK", Self->Items[i].Text, 0, STR_MATCHLENGTH) != ERR_Okay) {
               index = i;
            }
            break;
         }
      }
   }

   /*** Tool-tip management ***/

   if ((index != -1) AND (Self->Items[index].Tip)) {
      SubscribeTimer(Self->Head.UniqueID, 100);
   }

   /*** Do nothing if the selected item remains unchanged ***/

   if (index IS Self->Index) return(ERR_Okay);

   /*** Redraw the previous selected item ***/

   if (Self->Index != -1) {
      oldindex = Self->Index;
      Self->Index = -1;
      DrawItem(Self, oldindex);
   }

   /*** Draw the new selected index ***/

   Self->Index = index;
   if (index != -1) DrawItem(Self, index);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Bottom
Short:  The bottom coordinate of the toolbar.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The bottom coordinate of the toolbar (calculated as YCoord + Height) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Bottom(struct ToolBar *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Bottom, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  The drawable that will contain the toolbar graphic.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>The drawable that will contain the toolbar graphic is set here.  If this
field is not set prior to initialisation, the toolbar will attempt to scan
for the correct drawable by analysing its parents until it finds a suitable
candidate.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags may be set here.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Optional flags that may be set against a toolbar object are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">HIDE</td><td>This flag may be set on initialisation in order to hide the toolbar on its creation (applies to script usage only).</td></tr>
<tr><td width="10%">DISABLED</td><td>This readable flag is automatically set when the toolbar is disabled.  You may set it on initialisation if the toolbar is to be disabled on creation.</td></tr>
<tr><td width="10%">LOCAL</td><td>By default, the toolbar will send messages to its child objects when the toolbar state changes.  If you set the LOCAL option, the toolbar will execute the objects within its own process space.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a toolbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A toolbar can be given a fixed or relative height by setting this field
to the desired value.  To set a relative height, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetHeight, ERR_AccessObject));
}

static ERROR SET_Height(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Height, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   Highlight
Short:  String-based field for setting the toolbar highlight.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border highlight of a toolbar is set by writing to this field. The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of "#ff0000" or "255,0,0" would be valid.</p>

###SEE ALSO###
Fields: Colour, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Highlight(struct ToolBar *Self, STRING Value)
{
   if (Value) {
      StrReadColour(Value, &Self->HighlightRGB.Red, &Self->HighlightRGB.Green, &Self->HighlightRGB.Blue, &Self->HighlightRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HighlightRGB
Short:  Defines border highlight of the toolbar, in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The border highlight of the toolbar can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_HighlightRGB(struct ToolBar *Self, struct RGB *Value)
{
   Value->Red   = Self->HighlightRGB.Red;
   Value->Green = Self->HighlightRGB.Green;
   Value->Blue  = Self->HighlightRGB.Blue;
   Value->Alpha = Self->HighlightRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Right
Short:  The right coordinate of the toolbar.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The right coordinate of the toolbar (calculated as XCoord + Width) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Right(struct ToolBar *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Right, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   Shadow
Short:  String-based field for setting the toolbar shadow.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border shadow of a toolbar is set by writing to this field. The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of "#ff0000" or "255,0,0" would be valid.</p>

###SEE ALSO###
Fields: Colour, Highlight, Thickness

###END###

*****************************************************************************/

static ERROR SET_Shadow(struct ToolBar *Self, STRING Value)
{
   if (Value) {
      StrReadColour(Value, &Self->ShadowRGB.Red, &Self->ShadowRGB.Green, &Self->ShadowRGB.Blue, &Self->ShadowRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ShadowRGB
Short:  Defines border shadow of the toolbar, in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The border shadow of the toolbar can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ShadowRGB(struct ToolBar *Self, struct RGB *Value)
{
   Value->Red   = Self->ShadowRGB.Red;
   Value->Green = Self->ShadowRGB.Green;
   Value->Blue  = Self->ShadowRGB.Blue;
   Value->Alpha = Self->ShadowRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TabFocus
Short:  Setting this field to a valid TabFocus object will cause the toolbar to add itself to the tab list.
Type:   OBJECTID
Status: Set

###DESCRIPTION###
<p>The TabFocus field provides a convenient way of adding the toolbar to a
<a href="tabfocus.html">TabFocus</a> object, so that it can be focussed on via the tab key.  Simply set
this field to the ID of the TabFocus object that is managing the tab-list
for the application window.</p>

###SEE ALSO###
Class: TabFocus

###END###

*****************************************************************************/

static ERROR SET_TabFocus(struct ToolBar *Self, OBJECTID *Value)
{
   OBJECTPTR tabfocus;

   if (AccessObject(*Value, 5000, &tabfocus) IS ERR_Okay) {
      if (tabfocus->ClassID IS ID_TABFOCUS) {
         ActionTags(MT_AddTabObject, tabfocus, Self->RegionID);
      }
      ReleaseObject(tabfocus);
   }
   else return(ERR_AccessObject);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Template
Short:  Defines the makeup of the toolbar using a pre-defined template.
Type:   STRING
Status: Init

###DESCRIPTION###
<p>Setting the Template field allows a toolbar defintion to be loaded from a
pre-defined template file.  The template may contain settings for all of the
field values of a toolbar object, as well as a customised graphics script
for building the look and feel of the toolbar.  The template must be provided
as an XML file that can contain two possible root tags - 'value' and 'graphics'.
The value tag contains a list of field values for setting against the toolbar,
while the graphics tag contains the graphical makeup of the toolbar.  The
following example illustrates how a customised toolbar can be created:</p>

<pre>
&lt;?xml version="1.0"?&gt;

&lt;values&gt;
  &lt;exitframe value="1"/&gt;
  &lt;clickframe value="2"/&gt;
  &lt;enterframe value="3"/&gt;
  &lt;releaseframe value="1"/&gt;
&lt;/values&gt;

&lt;graphics&gt;
  &lt;box frame="1" boxes="(r0,0,!0,!0) (r1,1,!1,!1)" colour="230,230,230"
    highlight(0)="120,120,120" highlight(1)="255,255,255" shadow(0)="80,80,80"
    shadow(1)="200,200,200"/&gt;
&lt;/graphics&gt;
</pre>

###END###

*****************************************************************************/

static ERROR SET_Template(struct ToolBar *Self, STRING Value)
{
   struct XMLTag **xmllist, *tag, *child;
   UBYTE value[300];
   WORD i;

   if (Self->TemplateXML) {
      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   if (AnalyseLocation(Value, NULL) IS ERR_Okay) {
      if (CreateObject(ID_XML, NF_CHILD, &Self->TemplateXML, NULL,
         FID_Location|TSTRING, Value,
         TAGEND) IS ERR_Okay) {

         if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
            for (tag=xmllist[0]; tag; tag=tag->Next) {
               if (tag->Child) {
                  if (StrCompare("values", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     for (child=tag->Child; child; child=child->Next) {
                        if (child->TotalAttributes > 1) {
                           for (i=0; (i < sizeof(value)-1) AND (child->Attributes[1].Value[i]); i++) value[i] = child->Attributes[1].Value[i];
                           value[i] = 0;
                           StrTranslate(value, sizeof(value), NULL);
                           SetFieldVariable(Self, child->Attributes[0].Name, value);
                        }
                     }
                     break;
                  }
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Thickness
Short:  The thickness of the toolbar border.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the thickness of the toolbar border, expressed in
pixels.  By default the thickness is set to a value of 2.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a toolbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A toolbar can be given a fixed or relative width by setting this field
to the desired value.  To set a relative width, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Width(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetWidth, ERR_AccessObject));
}

static ERROR SET_Width(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Width, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of a toolbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of a toolbar can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetX, ERR_AccessObject));
}

static ERROR SET_X(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  The horizontal offset of a toolbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the toolbar will be
drawn from that X coordinate up to the width of the container, minus the
value given in the XOffset.  This means that the width of the ToolBar is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the toolbar will be drawn at an X coordinate calculated from the formula
&quot;XCoord = ContainerWidth - ToolBarWidth - XOffset&quot;.</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetXOffset, ERR_AccessObject));
}

static ERROR SET_XOffset(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical position of a toolbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of a ToolBar can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetY, ERR_AccessObject));

}

static ERROR SET_Y(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  The vertical offset of a toolbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the toolbar will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the toolbar is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the toolbar will be drawn at a Y coordinate calculated from the formula
&quot;YCoord = ContainerHeight - ToolBarHeight - YOffset&quot;.</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetYOffset, ERR_AccessObject));
}

static ERROR SET_YOffset(struct ToolBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************
** Internal: DrawToolBar()
*/

static void DrawToolBar(struct ToolBar *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct Bitmap *srcbitmap;
   struct Font *font;
   ULONG highlight, shadow, temp;
   WORD i, index, x, y, opacity;

   Self->DrawableWidth = Drawable->Width;
   Self->DrawableHeight = Drawable->Height;

   opacity = Bitmap->Opacity;

   /*** Draw the toolbar items ***/

   font = Self->Font;
   SetField(font, FID_Bitmap, FT_POINTER, Bitmap);

   for (index=0; index < Self->Total; index++) {
      x = Self->Items[index].XCoord;
      y = Self->Items[index].YCoord;

      if ((index IS Self->Index) AND (!(Drawable->Flags & RNF_DISABLED)) AND
          (!(Self->Items[index].Flags & TIF_DISABLED))){
         if (Self->Clicked) {
            /*** Darken the background when an item is clicked ***/
            Bitmap->Opacity = 32;
            ActionTags(MT_DrawRGBRectangle, Bitmap, x, y, Self->Items[index].Width, Self->Items[index].Height, 0, 0, 0, TRUE);
         }
         else {
            /*** Lighten the background when an item is selected ***/
            Bitmap->Opacity = 128;
            ActionTags(MT_DrawRGBRectangle, Bitmap, x, y, Self->Items[index].Width, Self->Items[index].Height, 255, 255, 255, TRUE);
         }

         /*** Draw the button border for the selected item ***/

         for (i=0; i < 2; i++) {
            /*** Top, Bottom ***/

            if (i > 0) Bitmap->Opacity = 255 - (255 / i);
            else Bitmap->Opacity = 255;

            shadow    = PackPixel(Bitmap, Self->ShadowRGB.Red, Self->ShadowRGB.Green, Self->ShadowRGB.Blue);
            highlight = PackPixel(Bitmap, Self->HighlightRGB.Red, Self->HighlightRGB.Green, Self->HighlightRGB.Blue);

            if (Self->Clicked) {
               temp      = highlight;
               highlight = shadow;
               shadow    = temp;
            }

            ActionTags(MT_DrawRectangle, Bitmap, x + i, y + i, Self->Items[index].Width-i-i, 1, highlight, TRUE);
            ActionTags(MT_DrawRectangle, Bitmap, x + i, y + Self->Items[index].Height-i-1, Self->Items[index].Width-i-i, 1, shadow, TRUE);

            /*** Left, Right ***/

            ActionTags(MT_DrawRectangle, Bitmap, x + i,                   y + i + 1, 1, Self->Items[index].Height-i-i-2, highlight, TRUE);
            ActionTags(MT_DrawRectangle, Bitmap, x + Self->Items[index].Width-i-1, y + i + 1, 1, Self->Items[index].Height-i-i-2, shadow, TRUE);
         }

         Bitmap->Opacity = 255;
      }

      if (StrCompare("BREAK", Self->Items[index].Text, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if (Self->Items[index].Height > Self->Items[index].Width) {
            ActionTags(MT_DrawRGBRectangle, Bitmap, x + (Self->Items[index].Width/2),   y+2, 1, Self->DrawableHeight - ((y + 2)*2), Self->ShadowRGB.Red, Self->ShadowRGB.Green, Self->ShadowRGB.Blue, TRUE);
            ActionTags(MT_DrawRGBRectangle, Bitmap, x + (Self->Items[index].Width/2)+1, y+2, 1, Self->DrawableHeight - ((y + 2)*2), Self->HighlightRGB.Red, Self->HighlightRGB.Green, Self->HighlightRGB.Blue, TRUE);
         }
         else {
            ActionTags(MT_DrawRGBRectangle, Bitmap, x, y, Self->Items[index].Width, 1, Self->ShadowRGB.Red, Self->ShadowRGB.Green, Self->ShadowRGB.Blue, TRUE);
            ActionTags(MT_DrawRGBRectangle, Bitmap, x, y+1, Self->Items[index].Width, 1, Self->HighlightRGB.Red, Self->HighlightRGB.Green, Self->HighlightRGB.Blue, TRUE);
         }
      }
      else if (Self->Items[index].Picture) {
         srcbitmap = Self->Items[index].Picture->Bitmap;

         if (srcbitmap->Flags & BMF_ALPHA) srcbitmap->Flags |= BMF_ALPHABLEND;

         if ((Drawable->Flags & RNF_DISABLED) OR (Self->Items[index].Flags & TIF_DISABLED)) srcbitmap->Opacity = 128;

         ActionTags(MT_BlitArea, srcbitmap, Bitmap, NULL, 0, 0, srcbitmap->Width, srcbitmap->Height,
            x + ((Self->Items[index].Width - srcbitmap->Width)/2), y + ((Self->Items[index].Height - srcbitmap->Height)/2));

         srcbitmap->Opacity = 255;
         srcbitmap->Flags &= ~BMF_ALPHABLEND;
      }
      else if (Self->Items[index].Text[0]) {
         SetField(font, FID_String, FT_STRING, Self->Items[index].Text);

         if ((Drawable->Flags & RNF_DISABLED) OR (Self->Items[index].Flags & TIF_DISABLED)) SetFields(font, FID_Opacity|TLONG, 25, TAGEND);

         font->XCoord = x;
         font->YCoord = Self->Thickness;
         font->AlignFlags |= ALIGN_VERTICAL;
         font->AlignHeight = Drawable->Height - Self->Thickness;
         Action(AC_Draw, font, NULL);

         if ((Drawable->Flags & RNF_DISABLED) OR (Self->Items[index].Flags & TIF_DISABLED)) SetFields(font, FID_Opacity|TLONG, 0, TAGEND);
      }

      Bitmap->Opacity = opacity;
   }

   if (!(Self->Flags & TBF_NOBKGD)) {

      shadow    = PackPixel(Bitmap, Self->ShadowRGB.Red, Self->ShadowRGB.Green, Self->ShadowRGB.Blue);
      highlight = PackPixel(Bitmap, Self->HighlightRGB.Red, Self->HighlightRGB.Green, Self->HighlightRGB.Blue);

      if (Self->Flags & TBF_SUNKEN) {
         /*** Reverse the border definitions in sunken mode ***/
         temp      = highlight;
         highlight = shadow;
         shadow    = temp;
      }

      if ((Self->Flags & TBF_EMBOSSED) AND (Self->Thickness)) Self->Thickness = 2;

      if (Self->Thickness > 0) {
         for (i=0; i < Self->Thickness; i++) {
            /*** Top, Bottom ***/

            ActionTags(MT_DrawRectangle, Bitmap, i, i, Drawable->Width-i-i, 1, highlight, TRUE);
            ActionTags(MT_DrawRectangle, Bitmap, i, Drawable->Height-i-1, Drawable->Width-i-i, 1, shadow, TRUE);

            /*** Left, Right ***/

            ActionTags(MT_DrawRectangle, Bitmap, i,                   i+1, 1, Drawable->Height-i-i-2, highlight, TRUE);
            ActionTags(MT_DrawRectangle, Bitmap, Drawable->Width-i-1, i+1, 1, Drawable->Height-i-i-2, shadow, TRUE);

            if (Self->Flags & TBF_EMBOSSED) {
               temp      = highlight;
               highlight = shadow;
               shadow    = temp;
            }
         }
      }
   }
}

/*****************************************************************************
** Internal: ArrangeItems()
*/

static void ArrangeItems(struct ToolBar *Self)
{
   WORD i, x, y, size;

   if (Self->DrawableWidth > Self->DrawableHeight) {
      /*** Toolbar is arranged horizontally ***/
      size = Self->DrawableHeight - (Self->Thickness * 2);
   }
   else {
      /*** Toolbar is arranged vertically ***/
      size = Self->DrawableWidth - (Self->Thickness * 2);
   }

   x = Self->Thickness;
   y = Self->Thickness;
   for (i=0; i < Self->Total; i++) {
      Self->Items[i].XCoord = x;
      Self->Items[i].YCoord = y;

      if (StrCompare("BREAK", Self->Items[i].Text, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if (Self->DrawableWidth > Self->DrawableHeight) {
            Self->Items[i].Width  = 2;
            Self->Items[i].Height = size;
         }
         else {
            Self->Items[i].Width  = size;
            Self->Items[i].Height = 2;
         }
      }
      else {
         if (Self->Items[i].Picture) {
            Self->Items[i].Width  = size;
            Self->Items[i].Height = size;
         }
         else if (Self->Items[i].Text) {
            struct mtCalcStringWidth calc;
            calc.String = Self->Items[i].Text;
            calc.Length = -1;
            Action(MT_CalcStringWidth, Self->Font, &calc);
            Self->Items[i].Width  = calc.Width;
            Self->Items[i].Height = size;
         }
      }

      if (Self->DrawableWidth > Self->DrawableHeight) {
         Self->Items[i].Width += HGAP;
         x += Self->Items[i].Width;
      }
      else {
         Self->Items[i].Height += HGAP;
         y += Self->Items[i].Height;
      }
   }
}

/*****************************************************************************
** Internal: DrawItem()
** Short:    Executes a draw for the item at a specific location.
*/

static void DrawItem(struct ToolBar *Self, WORD Index)
{
   struct acDraw draw;

   if ((Index >= 0) AND (Index < Self->Total)) {
      draw.XCoord = Self->Items[Index].XCoord;
      draw.YCoord = Self->Items[Index].YCoord;
      draw.Width  = Self->Items[Index].Width;
      draw.Height = Self->Items[Index].Height;
      ActionMsg(AC_Draw, Self->RegionID, &draw);
   }
}

